#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    BitsInteger,
    BitStruct,
    ByteSwapped,
    Bytewise,
    Const,
    Default,
    GreedyRange,
    GreedyString,
    Int8ul,
    Int16ul,
    Int24ub,
    Int32ul,
    PaddedString,
    Padding,
    Select,
    Struct,
    Switch,
    this,
)

from bluetooth_mesh.messages.util import DictAdapter, EnumAdapter, Opcode


class DebugOpcode(IntEnum):
    OPCODE = 0xF53601


class DebugSubOpcode(IntEnum):
    RSSI_THRESHOLD_GET = 0x00
    RSSI_THRESHOLD_SET = 0x01
    RSSI_THRESHOLD_STATUS = 0x02
    RADIO_TEST = 0x03
    TIMESLOT_TX_POWER_GET = 0x04
    TIMESLOT_TX_POWER_SET = 0x05
    TIMESLOT_TX_POWER_STATUS = 0x06
    SOFTDEVICE_TX_POWER_GET = 0x07
    SOFTDEVICE_TX_POWER_SET = 0x08
    SOFTDEVICE_TX_POWER_STATUS = 0x09
    UPTIME_GET = 0x0A
    UPTIME_STATUS = 0x0B
    LAST_SW_FAULT_GET = 0x0C
    LAST_SW_FAULT_CLEAR = 0x0D
    LAST_SW_FAULT_STATUS = 0x0E
    SYSTEM_STATS_GET = 0x0F
    SYSTEM_STATS_STATUS = 0x10
    LAST_MALLOC_FAULT_GET = 0x11
    LAST_MALLOC_FAULT_CLEAR = 0x12
    LAST_MALLOC_FAULT_STATUS = 0x13
    LAST_FDS_FAULT_GET = 0x14
    LAST_FDS_FAULT_CLEAR = 0x15
    LAST_FDS_FAULT_STATUS = 0x16
    BYTES_BEFORE_GARBAGE_COLLECTOR_GET = 0x17
    BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS = 0x18
    PROVISIONED_APP_VERSION_GET = 0x19
    PROVISIONED_APP_VERSION_STATUS = 0x1A
    FULL_FIRMWARE_VERSION_GET = 0x1B
    FULL_FIRMWARE_VERSION_STATUS = 0x1C
    IV_INDEX_GET = 0x1D
    IV_INDEX_STATUS = 0x1E
    GARBAGE_COLLECTOR_COUNTER_GET = 0x1F
    GARBAGE_COLLECTOR_COUNTER_STATUS = 0x20
    ARAP_LIST_SIZE_GET = 0x21
    ARAP_LIST_SIZE_STATUS = 0x22
    ARAP_LIST_CONTENT_GET = 0x23
    ARAP_LIST_CONTENT_STATUS = 0x24


# fmt: off
RssiThreshold = Struct(
    "rssi_threshold" / Int8ul
)

RadioTest = Struct(
    "packet_counter" / Int8ul
)

TxPower = Struct(
    "tx_power" / Int8ul
)

UptimeStatus = Struct(
    "uptime" / Int32ul
)

LastFault = Struct(
    "time" / Int32ul,
    "fault" / GreedyString("utf8")
)

SystemStats = Struct(
    "stats" / DictAdapter(
        GreedyRange(
            Struct(
                "name" / PaddedString(8, "utf8"),
                "high_water_mark" / Int16ul,
                "rfu" / Padding(4),
            )
        ),
        key=this.name,
        value=this.high_water_mark
    )
)

GarbageCollector = Struct(
    "bytes_left" / Int16ul
)

AppVersion = Struct(
    "version" / Int16ul
)

FirmwareVersion = Struct(
    "version" / GreedyString("utf8")
)

IvIndex = Struct(
    "ivindex" / Int32ul
)

GarbageCollectorCounter = Struct(
    "counter" / Int16ul
)

ArapSize = Select(
    Struct(
        "capacity" / Int16ul,
        "size" / Int16ul
    ),
    Struct(
        "capacity" / Int8ul,
        "size" / Int8ul
    )
)

ArapContentGet = Struct(
    "page" / Int8ul
)

ArapContent = Struct(
    "current_page" / Int8ul,
    "last_page" / Int8ul,
    "nodes" / DictAdapter(
        GreedyRange(
            ByteSwapped(
                BitStruct(
                    "sequence" / Bytewise(Int24ub),
                    "ivi" / BitsInteger(1),
                    "address" / BitsInteger(15),
                )
            )
        ),
        key=this.address,
        value=[this.ivi, this.sequence]
    )
)

DebugPayload = Struct(
    "subopcode" / EnumAdapter(Int8ul, DebugSubOpcode),
    "data" / Default(Switch(
        this.subopcode,
        {
            DebugSubOpcode.RSSI_THRESHOLD_SET: RssiThreshold,
            DebugSubOpcode.RSSI_THRESHOLD_STATUS: RssiThreshold,
            DebugSubOpcode.RADIO_TEST: RadioTest,
            DebugSubOpcode.TIMESLOT_TX_POWER_SET: TxPower,
            DebugSubOpcode.TIMESLOT_TX_POWER_STATUS: TxPower,
            DebugSubOpcode.SOFTDEVICE_TX_POWER_SET: TxPower,
            DebugSubOpcode.SOFTDEVICE_TX_POWER_STATUS: TxPower,
            DebugSubOpcode.UPTIME_STATUS: UptimeStatus,
            DebugSubOpcode.LAST_SW_FAULT_STATUS: LastFault,
            DebugSubOpcode.SYSTEM_STATS_STATUS: SystemStats,
            DebugSubOpcode.LAST_MALLOC_FAULT_STATUS: LastFault,
            DebugSubOpcode.LAST_FDS_FAULT_STATUS: LastFault,
            DebugSubOpcode.BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS: GarbageCollector,
            DebugSubOpcode.PROVISIONED_APP_VERSION_STATUS: AppVersion,
            DebugSubOpcode.FULL_FIRMWARE_VERSION_STATUS: FirmwareVersion,
            DebugSubOpcode.IV_INDEX_STATUS: IvIndex,
            DebugSubOpcode.GARBAGE_COLLECTOR_COUNTER_STATUS: GarbageCollectorCounter,
            DebugSubOpcode.ARAP_LIST_SIZE_STATUS: ArapSize,
            DebugSubOpcode.ARAP_LIST_CONTENT_GET: ArapContentGet,
            DebugSubOpcode.ARAP_LIST_CONTENT_STATUS: ArapContent,
        },
    ), None)
)

DebugMessage = Struct(
    "opcode" / Const(DebugOpcode.OPCODE, Opcode(DebugOpcode)),
    "params" / DebugPayload
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from asyncio import Future

import pytest

from bluetooth_mesh.utils import Gatherer


@pytest.mark.asyncio
async def test_gatherer():
    aws = [Future(), Future(), Future(), Future(), Future()]
    iaws = iter(aws)

    fut = next(iaws)
    fut.set_result(aws.index(fut))

    async for status in Gatherer(aws):
        assert status == (fut, aws.index(fut))
        fut = next(iaws, None)
        if fut:
            fut.set_result(aws.index(fut))


@pytest.mark.asyncio
async def test_gatherer_timeout():
    aws = [Future(), Future()]

    async for status in Gatherer(aws, timeout=0.1):
        key, val = status
        assert isinstance(val, TimeoutError)


@pytest.mark.asyncio
async def test_gatherer_exception():
    async def raises():
        raise Exception("ex")

    aws = [raises()]

    async for status in Gatherer(aws, timeout=0.1):
        stat, result = status
        assert stat == aws[0]
        assert isinstance(result, Exception)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=redefined-outer-name, invalid-name
import asynctest
from asynctest import mock, ANY

from bluetooth_mesh import LightLightnessClient, Element
from bluetooth_mesh.messages.config import GATTNamespaceDescriptor
from bluetooth_mesh.messages.generic import LightLightnessSetupMessage
from bluetooth_mesh.messages.generic.light import LightLightnessSetupOpcode
from bluetooth_mesh.test.fixtures import *  # pylint: disable=unused-wildcard-import, wildcard-import


class LLElementMock(Element):
    LOCATION = GATTNamespaceDescriptor.BOTTOM
    MODELS = [LightLightnessClient]


@pytest.fixture
def light_lightness_client(element_path) -> LightLightnessClient:
    element = LLElementMock(mock.MagicMock(), mock.MagicMock())
    element.path = element_path
    return LightLightnessClient(element)


@pytest.mark.asyncio
@asynctest.patch("asyncio.sleep", new=asynctest.CoroutineMock())
async def test_sending_scene_recall_repeated_6_times_with_intervals_by_default(
    light_lightness_client, destination, app_index
):
    light_lightness_client.repeat = asynctest.CoroutineMock()

    await light_lightness_client.set_lightness_range_unack(
        destination=destination, app_index=app_index, min_lightness=0, max_lightness=100
    )

    light_lightness_client.repeat.assert_awaited_once_with(
        ANY, retransmissions=6, send_interval=0.075
    )


# pylint: disable=protected-access
@pytest.mark.asyncio
@asynctest.patch("asyncio.sleep", new=asynctest.CoroutineMock())
async def test_scene_recall_calls_node_interface_with_appropriate_arguments(
    light_lightness_client, destination, app_index, element_path
):
    light_lightness_client._node_interface.send = asynctest.CoroutineMock()

    await light_lightness_client.set_lightness_range_unack(
        destination=destination, app_index=app_index, min_lightness=0, max_lightness=100
    )

    data = LightLightnessSetupMessage.build(
        {
            "opcode": LightLightnessSetupOpcode.LIGHTNESS_RANGE_SET_UNACKNOWLEDGED,
            "params": {"range_min": 0, "range_max": 100,},
        }
    )
    light_lightness_client._node_interface.send.assert_awaited_with(
        element_path, destination, app_index, data
    )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0223

import enum
from datetime import timedelta

from construct import (
    Adapter,
    BitsInteger,
    BitStruct,
    Bytes,
    Embedded,
    ExprValidator,
    Flag,
    GreedyBytes,
    GreedyRange,
    Int8ul,
    Int16ul,
    Int24ul,
    Padding,
    Rebuild,
    Select,
    Struct,
    Switch,
    len_,
    obj_,
    this,
)

from .util import (
    BitList,
    EmbeddedBitStruct,
    EnumAdapter,
    LogAdapter,
    Opcode,
    RangeValidator,
    Reversed,
)


class SecureNetworkBeacon(enum.IntEnum):
    OFF = 0x00
    ON = 0x01


class GATTProxy(enum.IntEnum):
    DISABLED = 0x00
    ENABLED = 0x01
    NOT_SUPPORTED = 0x02


class Relay(enum.IntEnum):
    DISABLED = 0x00
    ENABLED = 0x01
    NOT_SUPPORTED = 0x02


class GATTNamespaceDescriptor(enum.IntEnum):
    AUXILIARY = 0x0108
    BACK = 0x0101
    BACKUP = 0x0107
    BOTTOM = 0x0103
    EIGHTEENTH = 0x0012
    EIGHTH = 0x0008
    EIGHTIETH = 0x0050
    EIGHTY_EIGHTH = 0x0058
    EIGHTY_FIFTH = 0x0055
    EIGHTY_FIRST = 0x0051
    EIGHTY_FOURTH = 0x0054
    EIGHTY_NINTH = 0x0059
    EIGHTY_SECOND = 0x0052
    EIGHTY_SEVENTH = 0x0057
    EIGHTY_SIXTH = 0x0056
    EIGHTY_THIRD = 0x0053
    ELEVENTH = 0x000B
    EXTERNAL = 0x0110
    FIFTEENTH = 0x000F
    FIFTH = 0x0005
    FIFTIETH = 0x0032
    FIFTY_EIGHTH = 0x003A
    FIFTY_FIFTH = 0x0037
    FIFTY_FIRST = 0x0033
    FIFTY_FOURTH = 0x0036
    FIFTY_NINTH = 0x003B
    FIFTY_SECOND = 0x0034
    FIFTY_SEVENTH = 0x0039
    FIFTY_SIXTH = 0x0038
    FIFTY_THIRD = 0x0035
    FIRST = 0x0001
    FLASH = 0x010A
    FORTIETH = 0x0028
    FOURTEENTH = 0x000E
    FOURTH = 0x0004
    FORTY_EIGHTH = 0x0030
    FORTY_FIFTH = 0x002D
    FORTY_FIRST = 0x0029
    FORTY_FOURTH = 0x002C
    FORTY_NINTH = 0x0031
    FORTY_SECOND = 0x002A
    FORTY_SEVENTH = 0x002F
    FORTY_SIXTH = 0x002E
    FORTY_THIRD = 0x002B
    FRONT = 0x0100
    INSIDE = 0x010B
    INTERNAL = 0x010F
    LEFT = 0x010D
    LOWER = 0x0105
    MAIN = 0x0106
    NINETEENTH = 0x0013
    NINTH = 0x0009
    NINETIETH = 0x005A
    NINETY_EIGHTH = 0x0062
    NINETY_FIFTH = 0x005F
    NINETY_FIRST = 0x005B
    NINETY_FOURTH = 0x005E
    NINETY_NINTH = 0x0063
    NINETY_SECOND = 0x005C
    NINETY_SEVENTH = 0x0061
    NINETY_SIXTH = 0x0060
    NINETY_THIRD = 0x005D
    ONE_HUNDRED_AND_EIGHTEENTH = 0x0076
    ONE_HUNDRED_AND_EIGHTH = 0x006C
    ONE_HUNDRED_AND_EIGHTY_EIGHTH = 0x00BC
    ONE_HUNDRED_AND_EIGHTY_FIFTH = 0x00B9
    ONE_HUNDRED_AND_EIGHTY_FIRST = 0x00B5
    ONE_HUNDRED_AND_EIGHTY_FOURTH = 0x00B8
    ONE_HUNDRED_AND_EIGHTY_NINTH = 0x00BD
    ONE_HUNDRED_AND_EIGHTY_SECOND = 0x00B6
    ONE_HUNDRED_AND_EIGHTY_SEVENTH = 0x00BB
    ONE_HUNDRED_AND_EIGHTY_SIXTH = 0x00BA
    ONE_HUNDRED_AND_EIGHTY_THIRD = 0x00B7
    ONE_HUNDRED_AND_ELEVENTH = 0x006F
    ONE_HUNDRED_AND_FIFTEENTH = 0x0073
    ONE_HUNDRED_AND_FIFTH = 0x0069
    ONE_HUNDRED_AND_FIFTY_EIGHTH = 0x009E
    ONE_HUNDRED_AND_FIFTY_FIFTH = 0x009B
    ONE_HUNDRED_AND_FIFTY_FIRST = 0x0097
    ONE_HUNDRED_AND_FIFTY_FOURTH = 0x009A
    ONE_HUNDRED_AND_FIFTY_NINTH = 0x009F
    ONE_HUNDRED_AND_FIFTY_SECOND = 0x0098
    ONE_HUNDRED_AND_FIFTY_SEVENTH = 0x009D
    ONE_HUNDRED_AND_FIFTY_SIXTH = 0x009C
    ONE_HUNDRED_AND_FIFTY_THIRD = 0x0099
    ONE_HUNDRED_AND_FIRST = 0x0065
    ONE_HUNDRED_AND_FOURTEENTH = 0x0072
    ONE_HUNDRED_AND_FOURTH = 0x0068
    ONE_HUNDRED_AND_FORTY_EIGHTH = 0x0094
    ONE_HUNDRED_AND_FORTY_FIFTH = 0x0091
    ONE_HUNDRED_AND_FORTY_FIRST = 0x008D
    ONE_HUNDRED_AND_FORTY_FOURTH = 0x0090
    ONE_HUNDRED_AND_FORTY_NINTH = 0x0095
    ONE_HUNDRED_AND_FORTY_SECOND = 0x008E
    ONE_HUNDRED_AND_FORTY_SEVENTH = 0x0093
    ONE_HUNDRED_AND_FORTY_SIXTH = 0x0092
    ONE_HUNDRED_AND_FORTY_THIRD = 0x008F
    ONE_HUNDRED_AND_NINETEENTH = 0x0077
    ONE_HUNDRED_AND_NINTH = 0x006D
    ONE_HUNDRED_AND_NINETY_EIGHTH = 0x00C6
    ONE_HUNDRED_AND_NINETY_FIFTH = 0x00C3
    ONE_HUNDRED_AND_NINETY_FIRST = 0x00BF
    ONE_HUNDRED_AND_NINETY_FOURTH = 0x00C2
    ONE_HUNDRED_AND_NINETY_NINTH = 0x00C7
    ONE_HUNDRED_AND_NINETY_SECOND = 0x00C0
    ONE_HUNDRED_AND_NINETY_SEVENTH = 0x00C5
    ONE_HUNDRED_AND_NINETY_SIXTH = 0x00C4
    ONE_HUNDRED_AND_NINETY_THIRD = 0x00C1
    ONE_HUNDRED_AND_SECOND = 0x0066
    ONE_HUNDRED_AND_SEVENTEENTH = 0x0075
    ONE_HUNDRED_AND_SEVENTH = 0x006B
    ONE_HUNDRED_AND_SEVENTY_EIGHTH = 0x00B2
    ONE_HUNDRED_AND_SEVENTY_FIFTH = 0x00AF
    ONE_HUNDRED_AND_SEVENTY_FIRST = 0x00AB
    ONE_HUNDRED_AND_SEVENTY_FOURTH = 0x00AE
    ONE_HUNDRED_AND_SEVENTY_NINTH = 0x00B3
    ONE_HUNDRED_AND_SEVENTY_SECOND = 0x00AC
    ONE_HUNDRED_AND_SEVENTY_SEVENTH = 0x00B1
    ONE_HUNDRED_AND_SEVENTY_SIXTH = 0x00B0
    ONE_HUNDRED_AND_SEVENTY_THIRD = 0x00AD
    ONE_HUNDRED_AND_SIXTEENTH = 0x0074
    ONE_HUNDRED_AND_SIXTH = 0x006A
    ONE_HUNDRED_AND_SIXTY_EIGHTH = 0x00A8
    ONE_HUNDRED_AND_SIXTY_FIFTH = 0x00A5
    ONE_HUNDRED_AND_SIXTY_FIRST = 0x00A1
    ONE_HUNDRED_AND_SIXTY_FOURTH = 0x00A4
    ONE_HUNDRED_AND_SIXTY_NINTH = 0x00A9
    ONE_HUNDRED_AND_SIXTY_SECOND = 0x00A2
    ONE_HUNDRED_AND_SIXTY_SEVENTH = 0x00A7
    ONE_HUNDRED_AND_SIXTY_SIXTH = 0x00A6
    ONE_HUNDRED_AND_SIXTY_THIRD = 0x00A3
    ONE_HUNDRED_AND_TENTH = 0x006E
    ONE_HUNDRED_AND_THIRD = 0x0067
    ONE_HUNDRED_AND_THIRTEENTH = 0x0071
    ONE_HUNDRED_AND_THIRTY_EIGHTH = 0x008A
    ONE_HUNDRED_AND_THIRTY_FIFTH = 0x0087
    ONE_HUNDRED_AND_THIRTY_FIRST = 0x0083
    ONE_HUNDRED_AND_THIRTY_FOURTH = 0x0086
    ONE_HUNDRED_AND_THIRTY_NINTH = 0x008B
    ONE_HUNDRED_AND_THIRTY_SECOND = 0x0084
    ONE_HUNDRED_AND_THIRTY_SEVENTH = 0x0089
    ONE_HUNDRED_AND_THIRTY_SIXTH = 0x0088
    ONE_HUNDRED_AND_THIRTY_THIRD = 0x0085
    ONE_HUNDRED_AND_TWELVETH = 0x0070
    ONE_HUNDRED_AND_TWENTY_EIGHTH = 0x0080
    ONE_HUNDRED_AND_TWENTY_FIFTH = 0x007D
    ONE_HUNDRED_AND_TWENTY_FIRST = 0x0079
    ONE_HUNDRED_AND_TWENTY_FOURTH = 0x007C
    ONE_HUNDRED_AND_TWENTY_NINTH = 0x0081
    ONE_HUNDRED_AND_TWENTY_SECOND = 0x007A
    ONE_HUNDRED_AND_TWENTY_SEVENTH = 0x007F
    ONE_HUNDRED_AND_TWENTY_SIXTH = 0x007E
    ONE_HUNDRED_AND_TWENTY_THIRD = 0x007B
    ONE_HUNDRED_EIGHTIETH = 0x00B4
    ONE_HUNDRED_FIFTIETH = 0x0096
    ONE_HUNDRED_FORTIETH = 0x008C
    ONE_HUNDRED_NINETIETH = 0x00BE
    ONE_HUNDRED_SEVENTIETH = 0x00AA
    ONE_HUNDRED_SIXTIETH = 0x00A0
    ONE_HUNDRED_THIRTIETH = 0x0082
    ONE_HUNDRED_TWENTIETH = 0x0078
    ONE_HUNDREDTH = 0x0064
    OUTSIDE = 0x010C
    RIGHT = 0x010E
    SECOND = 0x0002
    SEVENTEENTH = 0x0011
    SEVENTH = 0x0007
    SEVENTIETH = 0x0046
    SEVENTY_EIGHTH = 0x004E
    SEVENTY_FIFTH = 0x004B
    SEVENTY_FIRST = 0x0047
    SEVENTY_FOURTH = 0x004A
    SEVENTY_NINTH = 0x004F
    SEVENTY_SECOND = 0x0048
    SEVENTY_SEVENTH = 0x004D
    SEVENTY_SIXTH = 0x004C
    SEVENTY_THIRD = 0x0049
    SIXTEENTH = 0x0010
    SIXTH = 0x0006
    SIXTIETH = 0x003C
    SIXTY_EIGHTH = 0x0044
    SIXTY_FIFTH = 0x0041
    SIXTY_FIRST = 0x003D
    SIXTY_FOURTH = 0x0040
    SIXTY_NINTH = 0x0045
    SIXTY_SECOND = 0x003E
    SIXTY_SEVENTH = 0x0043
    SIXTY_SIXTH = 0x0042
    SIXTY_THIRD = 0x003F
    SUPPLEMENTARY = 0x0109
    TENTH = 0x000A
    THIRD = 0x0003
    THIRTEENTH = 0x000D
    THIRTIETH = 0x001E
    THIRTY_EIGHTH = 0x0026
    THIRTY_FIFTH = 0x0023
    THIRTY_FIRST = 0x001F
    THIRTY_FOURTH = 0x0022
    THIRTY_NINTH = 0x0027
    THIRTY_SECOND = 0x0020
    THIRTY_SEVENTH = 0x0025
    THIRTY_SIXTH = 0x0024
    THIRTY_THIRD = 0x0021
    TOP = 0x0102
    TWELVETH = 0x000C
    TWENTIETH = 0x0014
    TWENTY_EIGHTH = 0x001C
    TWENTY_FIFTH = 0x0019
    TWENTY_FIRST = 0x0015
    TWENTY_FOURTH = 0x0018
    TWENTY_NINTH = 0x001D
    TWENTY_SECOND = 0x0016
    TWENTY_SEVENTH = 0x001B
    TWENTY_SIXTH = 0x001A
    TWENTY_THIRD = 0x0017
    TWO_HUNDRED_AND_EIGHTEENTH = 0x00DA
    TWO_HUNDRED_AND_EIGHTH = 0x00D0
    TWO_HUNDRED_AND_ELEVENTH = 0x00D3
    TWO_HUNDRED_AND_FIFTEENTH = 0x00D7
    TWO_HUNDRED_AND_FIFTH = 0x00CD
    TWO_HUNDRED_AND_FIFTY_FIFTH = 0x00FF
    TWO_HUNDRED_AND_FIFTY_FIRST = 0x00FB
    TWO_HUNDRED_AND_FIFTY_FOURTH = 0x00FE
    TWO_HUNDRED_AND_FIFTY_SECOND = 0x00FC
    TWO_HUNDRED_AND_FIFTY_THIRD = 0x00FD
    TWO_HUNDRED_AND_FIRST = 0x00C9
    TWO_HUNDRED_AND_FOURTEENTH = 0x00D6
    TWO_HUNDRED_AND_FOURTH = 0x00CC
    TWO_HUNDRED_AND_FORTY_EIGHTH = 0x00F8
    TWO_HUNDRED_AND_FORTY_FIFTH = 0x00F5
    TWO_HUNDRED_AND_FORTY_FIRST = 0x00F1
    TWO_HUNDRED_AND_FORTY_FOURTH = 0x00F4
    TWO_HUNDRED_AND_FORTY_NINTH = 0x00F9
    TWO_HUNDRED_AND_FORTY_SECOND = 0x00F2
    TWO_HUNDRED_AND_FORTY_SEVENTH = 0x00F7
    TWO_HUNDRED_AND_FORTY_SIXTH = 0x00F6
    TWO_HUNDRED_AND_FORTY_THIRD = 0x00F3
    TWO_HUNDRED_AND_NINETEENTH = 0x00DB
    TWO_HUNDRED_AND_NINTH = 0x00D1
    TWO_HUNDRED_AND_SECOND = 0x00CA
    TWO_HUNDRED_AND_SEVENTEENTH = 0x00D9
    TWO_HUNDRED_AND_SEVENTH = 0x00CF
    TWO_HUNDRED_AND_SIXTEENTH = 0x00D8
    TWO_HUNDRED_AND_SIXTH = 0x00CE
    TWO_HUNDRED_AND_TENTH = 0x00D2
    TWO_HUNDRED_AND_THIRD = 0x00CB
    TWO_HUNDRED_AND_THIRTEENTH = 0x00D5
    TWO_HUNDRED_AND_THIRTY_EIGHTH = 0x00EE
    TWO_HUNDRED_AND_THIRTY_FIFTH = 0x00EB
    TWO_HUNDRED_AND_THIRTY_FIRST = 0x00E7
    TWO_HUNDRED_AND_THIRTY_FOURTH = 0x00EA
    TWO_HUNDRED_AND_THIRTY_NINTH = 0x00EF
    TWO_HUNDRED_AND_THIRTY_SECOND = 0x00E8
    TWO_HUNDRED_AND_THIRTY_SEVENTH = 0x00ED
    TWO_HUNDRED_AND_THIRTY_SIXTH = 0x00EC
    TWO_HUNDRED_AND_THIRTY_THIRD = 0x00E9
    TWO_HUNDRED_AND_TWELVETH = 0x00D4
    TWO_HUNDRED_AND_TWENTY_EIGHTH = 0x00E4
    TWO_HUNDRED_AND_TWENTY_FIFTH = 0x00E1
    TWO_HUNDRED_AND_TWENTY_FIRST = 0x00DD
    TWO_HUNDRED_AND_TWENTY_FOURTH = 0x00E0
    TWO_HUNDRED_AND_TWENTY_NINTH = 0x00E5
    TWO_HUNDRED_AND_TWENTY_SECOND = 0x00DE
    TWO_HUNDRED_AND_TWENTY_SEVENTH = 0x00E3
    TWO_HUNDRED_AND_TWENTY_SIXTH = 0x00E2
    TWO_HUNDRED_AND_TWENTY_THIRD = 0x00DF
    TWO_HUNDRED_FIFTIETH = 0x00FA
    TWO_HUNDRED_FORTIETH = 0x00F0
    TWO_HUNDRED_THIRTIETH = 0x00E6
    TWO_HUNDRED_TWENTIETH = 0x00DC
    TWO_HUNDREDTH = 0x00C8
    UNKNOWN = 0x0000
    UPPER = 0x0104


class PublishFriendshipCredentialsFlag(enum.IntEnum):
    MASTER_SECURITY = 0x00
    FRIENDSHIP_SECURITY = 0x01


class StatusCode(enum.IntEnum):
    SUCCESS = 0x00
    INVALID_ADDRESS = 0x01
    INVALID_MODEL = 0x02
    INVALID_APPKEY_INDEX = 0x03
    INVALID_NETKEY_INDEX = 0x04
    INSUFFICIENT_RESOURCES = 0x05
    KEY_INDEX_ALREADY_STORED = 0x06
    INVALID_PUBLISH_PARAMETERS = 0x07
    NOT_A_SUBSCRIBE_MODEL = 0x08
    STORAGE_FAILURE = 0x09
    FEATURE_NOT_SUPPORTED = 0x0A
    CANNOT_UPDATE = 0x0B
    CANNOT_REMOVE = 0x0C
    CANNOT_BIND = 0x0D
    TEMPORARILY_UNABLE_TO_CHANGE_STATE = 0x0E
    CANNOT_SET = 0x0F
    UNSPECIFIED_ERROR = 0x10
    INVALID_BINDING = 0x11


class NodeIdentity(enum.IntEnum):
    STOPPED = 0x00
    RUNNING = 0x01
    NOT_SUPPORTED = 0x02


class Friend(enum.IntEnum):
    DISABLED = 0x00
    ENABLED = 0x01
    NOT_SUPPORTED = 0x02


class KeyRefreshPhase(enum.IntEnum):
    NORMAL = 0x00
    FIRST = 0x01
    SECOND = 0x02


class KeyRefreshTransition(enum.IntEnum):
    SECOND = 0x02
    THIRD = 0x03


SecureNetworkBeaconAdapter = EnumAdapter(Int8ul, SecureNetworkBeacon)

GATTProxyAdapter = EnumAdapter(Int8ul, GATTProxy)

RelayAdapter = EnumAdapter(Int8ul, Relay)

GATTNamespaceDescriptorAdapter = EnumAdapter(Int16ul, GATTNamespaceDescriptor)

PublishFriendshipCredentialsFlagAdapter = EnumAdapter(
    Flag, PublishFriendshipCredentialsFlag
)

StatusCodeAdapter = EnumAdapter(Int8ul, StatusCode)

NodeIdentityAdapter = EnumAdapter(Int8ul, NodeIdentity)

FriendAdapter = EnumAdapter(Int8ul, Friend)

TTL = RangeValidator(Int8ul, max_value=0x7F)

# fmt: off
SIGModelId = Struct(
    "model_id" / Int16ul,
)

VendorModelId = Struct(
    "vendor_id" / Int16ul,
    "model_id" / Int16ul,
)

ModelId = Select(
    VendorModelId,
    SIGModelId
)
# fmt: on


class AddressType(enum.Enum):
    UNASSIGNED = 0
    UNICAST = 1
    GROUP = 2
    VIRTUAL = 3
    RFU = 4
    ALL_PROXIES = 5
    ALL_FRIENDS = 6
    ALL_RELAYS = 7
    ALL_NODES = 8


def get_address_type(address):
    # pylint: disable=R0911

    if address == 0x0000:
        return AddressType.UNASSIGNED

    if 0xFF00 <= address <= 0xFFFB:
        return AddressType.RFU

    if address == 0xFFFC:
        return AddressType.ALL_PROXIES

    if address == 0xFFFD:
        return AddressType.ALL_FRIENDS

    if address == 0xFFFE:
        return AddressType.ALL_RELAYS

    if address == 0xFFFF:
        return AddressType.ALL_NODES

    if (address & 0xC000) == 0xC000:
        return AddressType.GROUP

    if (address & 0x8000) == 0x8000:
        return AddressType.VIRTUAL

    return AddressType.UNICAST


def AddressTypeValidator(subcons, *allowed_types):
    return ExprValidator(
        subcons, lambda obj, ctx: get_address_type(obj) in allowed_types
    )


# fmt: off
UnassignedAddress = AddressTypeValidator(
    Int16ul,
    AddressType.UNASSIGNED
)

UnicastAddress = AddressTypeValidator(
    Int16ul,
    AddressType.UNICAST
)

# TODO Fixed Groups?
GroupAddress = AddressTypeValidator(
    Int16ul, AddressType.GROUP
)

VirtualLabel = AddressTypeValidator(
    Int16ul,
    AddressType.VIRTUAL
)

NotVirtualLabel = ExprValidator(
    Int16ul,
    lambda obj, ctx: get_address_type(obj) != AddressType.VIRTUAL
)

SubscriptionAddress = ExprValidator(
    Int16ul,
    lambda obj, ctx: get_address_type(obj) not in [AddressType.UNASSIGNED,
                                                   AddressType.UNICAST,
                                                   AddressType.ALL_NODES,
                                                   AddressType.VIRTUAL]
)

StatusSubscriptionAddress = ExprValidator(
    Int16ul,
    lambda obj, ctx: get_address_type(obj) not in [AddressType.UNICAST,
                                                   AddressType.ALL_NODES]
)

UnicastUnassignedAddress = AddressTypeValidator(
    Int16ul,
    AddressType.UNICAST,
    AddressType.UNASSIGNED,
)

UnicastUnassignedGroupAddress = AddressTypeValidator(
    Int16ul,
    AddressType.UNICAST,
    AddressType.UNASSIGNED,
    AddressType.GROUP,
)

Int12ul = ExprValidator(
    Int16ul,
    (obj_ & 0xF000) == 0x00
)

CompositionDataElement = Struct(
    "location" / GATTNamespaceDescriptorAdapter,
    "SIG_number" / Rebuild(Int8ul, len_(this["SIG_models"])),
    "vendor_number" / Rebuild(Int8ul, len_(this["vendor_models"])),
    "SIG_models" / SIGModelId[this["SIG_number"]],
    "vendor_models" / VendorModelId[this["vendor_number"]],
)

CompositionData = Struct(
    "CID" / Int16ul,
    "PID" / Int16ul,
    "VID" / Int16ul,
    "CRPL" / Int16ul,
    "features" / Int16ul,  # TODO should be parsed
    "elements" / GreedyRange(CompositionDataElement),
)

Retransmit = BitStruct(
    # sssssccc
    "interval_steps" / BitsInteger(5),
    "count" / BitsInteger(3),
)
# fmt: on


class RetransmitAdapter(Adapter):
    def __init__(self, subcon, interval):
        self.interval = interval
        self.subcon = subcon
        super(RetransmitAdapter, self).__init__(subcon)

    def _decode(self, obj, context, path):
        return dict(
            count=obj["count"], interval=(obj["interval_steps"] + 1) * self.interval
        )

    def _encode(self, obj, context, path):
        if obj["count"] > 7:
            raise AttributeError("Retransmission count maximum value is 7")
        if (
            not 0 < obj["interval"] <= 0x20 * self.interval
            or obj["interval"] % self.interval
        ):
            raise AttributeError(
                "Interval value must be in range of {min} and {max} with increment of {min}".format(
                    min=self.interval, max=0x20 * self.interval
                )
            )
        return dict(
            count=obj["count"],
            interval_steps=int(round((obj["interval"] / self.interval) - 1)),
        )


NetworkRetransmit = RetransmitAdapter(
    Retransmit, 10
)  # (Network Retransmit Interval Steps + 1) * 10ms

RelayRetransmit = NetworkRetransmit  # (Relay Retransmit Interval Steps + 1) * 10ms

PublishRetransmit = RetransmitAdapter(
    Retransmit, 50
)  # (Publish Retransmit Interval Steps + 1) * 50ms


def DoubleKeyIndex(first, second):
    return EmbeddedBitStruct(
        "_", second / BitsInteger(12), first / BitsInteger(12), reversed=True
    )


def SingleKeyIndex(name):
    return EmbeddedBitStruct("_", Padding(4), name / BitsInteger(12), reversed=True)


NetAndAppKeyIndex = DoubleKeyIndex("net_key_index", "app_key_index")

AppKeyIndex = SingleKeyIndex("app_key_index")

NetKeyIndex = SingleKeyIndex("net_key_index")


class KeyIndicesAdapter(Adapter):
    def _decode(self, obj, context, path):
        """
        Flatten a list dictionaries into list of items:
            [{first=1, second=2}, {last=3}] -> [1, 2, 3]
        """
        ret = []
        for item in obj:
            if "last" in item:
                ret += [item["last"]]
            else:
                ret += [item["first"], item["second"]]
        return sorted(ret)

    def _encode(self, obj, context, path):
        """
        Expand a list into list of dictionaries:
            [1, 2, 3] -> [{first=1, second=2}, {last=3}]
        """
        ret = []
        obj.sort()
        while len(obj) > 1:
            ret += [dict(first=obj.pop(0), second=obj.pop(0))]

        if obj:
            ret += [dict(last=obj.pop())]

        return ret


# fmt: off
KeyIndices = KeyIndicesAdapter(
    GreedyRange(
        Select(
            Reversed(
                BitStruct(
                    "first" / BitsInteger(12),
                    "second" / BitsInteger(12),
                )
            ),
            Reversed(
                BitStruct(
                    Padding(4),
                    "last" / BitsInteger(12),
                )
            )
        )
    )
)
KeyIndices.__construct_doc__ = GreedyRange(BitsInteger(12))
# fmt: on


class PublishPeriodStepResolution(enum.IntEnum):
    RESOLUTION_100_MS = 0x0
    RESOLUTION_1_S = 0x1
    RESOLUTION_10_S = 0x2
    RESOLUTION_10_MIN = 0x3

    @property
    def multiplier(self):
        if self == PublishPeriodStepResolution.RESOLUTION_100_MS:
            return timedelta(milliseconds=100)

        if self == PublishPeriodStepResolution.RESOLUTION_1_S:
            return timedelta(seconds=1)

        if self == PublishPeriodStepResolution.RESOLUTION_10_S:
            return timedelta(seconds=10)

        if self == PublishPeriodStepResolution.RESOLUTION_10_MIN:
            return timedelta(minutes=10)


# fmt: off
PublishPeriodStepResolutionAdapter = EnumAdapter(
    BitsInteger(2),
    PublishPeriodStepResolution
)

PublishPeriod = BitStruct(
    "step_resolution" / PublishPeriodStepResolutionAdapter,
    "number_of_steps" / BitsInteger(6),
)

ConfigBeaconGet = Struct()

ConfigBeaconSet = Struct(
    "beacon" / SecureNetworkBeaconAdapter,
)

ConfigBeaconStatus = ConfigBeaconSet

ConfigCompositionDataGet = Struct(
    "page" / Int8ul,
)

ConfigCompositionDataStatus = Struct(
    "page" / Int8ul,
    "data" / Switch(
        this.page,
        {
            0: CompositionData,
        },
        default=GreedyBytes
    ),
)

ConfigDefaultTTLGet = Struct()

ConfigDefaultTTLSet = Struct(
    "TTL" / TTL,
)

ConfigDefaultTTLStatus = ConfigDefaultTTLSet

ConfigGATTProxyGet = Struct()

ConfigGATTProxySet = Struct(
    "GATT_proxy" / GATTProxyAdapter,
)

ConfigGATTProxyStatus = ConfigGATTProxySet

ConfigRelayGet = Struct()

ConfigRelaySet = Struct(
    "relay" / RelayAdapter,
    "retransmit" / RelayRetransmit,
)

ConfigRelayStatus = ConfigRelaySet

ConfigModelPublicationGet = Struct(
    "element_address" / UnicastAddress,
    "model" / ModelId,
)

ConfigModelPublicationSet = Struct(
    "element_address" / UnicastAddress,
    "publish_address" / NotVirtualLabel,
    *EmbeddedBitStruct(
        "_",
        "RFU" / BitsInteger(3),
        "credential_flag" / PublishFriendshipCredentialsFlagAdapter,
        "app_key_index" / BitsInteger(12),
        reversed=True
    ),
    "TTL" / TTL,
    "publish_period" / PublishPeriod,
    "retransmit" / PublishRetransmit,
    "model" / ModelId,
)

ConfigModelPublicationStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigModelPublicationSet)
)

ConfigModelPublicationVASet = Struct(
    "element_address" / UnicastAddress,
    "publish_address" / Bytes(16),
    *EmbeddedBitStruct(
        "_",
        "RFU" / BitsInteger(3),
        "credential_flag" / PublishFriendshipCredentialsFlagAdapter,
        "app_key_index" / BitsInteger(12),
        reversed=True
    ),
    "TTL" / TTL,
    "publish_period" / PublishPeriod,
    "retransmit" / PublishRetransmit,
    "model" / ModelId,
)

ConfigModelSubscriptionAdd = Struct(
    "element_address" / UnicastAddress,
    "address" / SubscriptionAddress,
    "model" / ModelId,
)

ConfigModelSubscriptionDelete = ConfigModelSubscriptionAdd
ConfigModelSubscriptionOverwrite = ConfigModelSubscriptionAdd

ConfigModelSubscriptionVAAdd = Struct(
    "element_address" / UnicastAddress,
    "label" / Bytes(16),
    "model" / ModelId,
)

ConfigModelSubscriptionVADelete = ConfigModelSubscriptionVAAdd
ConfigModelSubscriptionVAOverwrite = ConfigModelSubscriptionVAAdd

ConfigModelSubscriptionStatus = Struct(
    "status" / StatusCodeAdapter,
    "element_address" / UnicastAddress,
    "address" / StatusSubscriptionAddress,
    "model" / ModelId,
)

ConfigModelSubscriptionDeleteAll = Struct(
    "element_address" / UnicastAddress,
    "model" / ModelId,
)

ConfigSIGModelSubscriptionGet = Struct(
    "element_address" / UnicastAddress,
    "model" / SIGModelId,
)

ConfigSIGModelSubscriptionList = Struct(
    "status" / StatusCodeAdapter,
    "element_address" / UnicastAddress,
    "model" / SIGModelId,
    "addresses" / GreedyRange(Int16ul),
)

ConfigVendorModelSubscriptionGet = Struct(
    "element_address" / UnicastAddress,
    "model" / VendorModelId,
)

ConfigVendorModelSubscriptionList = Struct(
    "status" / StatusCodeAdapter,
    "element_address" / UnicastAddress,
    "model" / VendorModelId,
    "addresses" / GreedyRange(Int16ul),
)

ConfigNetKeyAdd = Struct(
    *NetKeyIndex,
    "net_key" / Bytes(16),
)

ConfigNetKeyUpdate = ConfigNetKeyAdd

ConfigNetKeyDelete = Struct(
    *NetKeyIndex,
)

ConfigNetKeyStatus = Struct(
    "status" / StatusCodeAdapter,
    *NetKeyIndex,
)

ConfigNetKeyGet = Struct()

ConfigNetKeyList = Struct(
    "net_key_indices" / KeyIndices
)

ConfigAppKeyAdd = Struct(
    *NetAndAppKeyIndex,
    "app_key" / Bytes(16),
)

ConfigAppKeyUpdate = ConfigAppKeyAdd

ConfigAppKeyDelete = Struct(
    *NetAndAppKeyIndex,
)

ConfigAppKeyStatus = Struct(
    "status" / StatusCodeAdapter,
    *NetAndAppKeyIndex,
)

ConfigAppKeyGet = Struct(
    *NetKeyIndex
)

ConfigAppKeyList = Struct(
    "status" / StatusCodeAdapter,
    *NetKeyIndex,
    "app_key_indices" / KeyIndices,
)

ConfigNodeIdentityGet = Struct(
    *NetKeyIndex,
)

ConfigNodeIdentitySet = Struct(
    *NetKeyIndex,
    "identity" / NodeIdentityAdapter,
)

ConfigNodeIdentityStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigNodeIdentitySet),
)

ConfigModelAppBind = Struct(
    "element_address" / UnicastAddress,
    *AppKeyIndex,
    "model" / ModelId,
)

ConfigModelAppUnbind = ConfigModelAppBind

ConfigModelAppStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigModelAppBind),
)

ConfigSIGModelAppGet = Struct(
    "element_address" / UnicastAddress,
    "model" / SIGModelId,
)

ConfigSIGModelAppList = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigSIGModelAppGet),
    "app key indices" / KeyIndices,
)

ConfigVendorModelAppGet = Struct(
    "element_address" / UnicastAddress,
    "model" / VendorModelId,
)

ConfigVendorModelAppList = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigVendorModelAppGet),
    "app key indices" / KeyIndices,
)

ConfigNodeReset = Struct()

ConfigNodeResetStatus = Struct()

ConfigFriendGet = Struct()

ConfigFriendSet = Struct(
    "friend" / FriendAdapter,
)

ConfigFriendStatus = ConfigFriendSet

ConfigKeyRefreshPhaseGet = Struct(
    *NetKeyIndex,
)

ConfigKeyRefreshPhaseSet = Struct(
    Embedded(ConfigKeyRefreshPhaseGet),
    "transition" / EnumAdapter(Int8ul, KeyRefreshTransition),
)

ConfigKeyRefreshPhaseStatus = Struct(
    "status" / StatusCodeAdapter,
    *NetKeyIndex,
    "phase" / EnumAdapter(Int8ul, KeyRefreshPhase),
)

ConfigHeartbeatPublicationFeatures = ExprValidator(Int16ul, lambda obj, ctx: obj <= 0x3)

ConfigHeartbeatHops = ExprValidator(Int8ul, lambda obj, ctx: obj <= 0x7F)

ConfigHeartbeatPublicationGet = Struct()

ConfigHeartbeatPublicationSet = Struct(
    "destination" / UnicastUnassignedGroupAddress,
    "count" / LogAdapter(Int8ul, max_value=0x10, infinity=True),
    "period" / LogAdapter(Int8ul, max_value=0x10),
    "TTL" / TTL,
    "features" / BitList(2),
    *NetKeyIndex,
)

ConfigHeartbeatPublicationStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigHeartbeatPublicationSet),
)

ConfigHeartbeatSubscriptionGet = Struct()

ConfigHeartbeatSubscriptionSet = Struct(
    "source" / UnicastUnassignedAddress,
    "destination" / UnicastUnassignedGroupAddress,
    "period_log" / LogAdapter(Int8ul, max_value=0x11),
)

ConfigHeartbeatSubscriptionStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(ConfigHeartbeatSubscriptionSet),
    "count" / LogAdapter(Int8ul, max_value=0x11, infinity=True),
    "min_hops" / RangeValidator(Int8ul, max_value=0x7F),
    "max_hops" / RangeValidator(Int8ul, max_value=0x7F),
)

ConfigLowPowerNodePollTimeoutGet = Struct(
    "lpn_address" / Int16ul,  # TODO
)

ConfigLowPowerNodePollTimeoutStatus = Struct(
    Embedded(ConfigLowPowerNodePollTimeoutGet),
    "poll_timeout" / Int24ul,  # TODO
)

ConfigNetworkTransmitGet = Struct()

ConfigNetworkTransmitSet = NetworkRetransmit

ConfigNetworkTransmitStatus = ConfigNetworkTransmitSet
# fmt: on


class ConfigOpcode(enum.IntEnum):
    APPKEY_ADD = 0x00
    APPKEY_DELETE = 0x8000
    APPKEY_GET = 0x8001
    APPKEY_LIST = 0x8002
    APPKEY_STATUS = 0x8003
    APPKEY_UPDATE = 0x01
    BEACON_GET = 0x8009
    BEACON_SET = 0x800A
    BEACON_STATUS = 0x800B
    COMPOSITION_DATA_GET = 0x8008
    COMPOSITION_DATA_STATUS = 0x02
    DEFAULT_TTL_GET = 0x800C
    DEFAULT_TTL_SET = 0x800D
    DEFAULT_TTL_STATUS = 0x800E
    FRIEND_GET = 0x800F
    FRIEND_SET = 0x8010
    FRIEND_STATUS = 0x8011
    GATT_PROXY_GET = 0x8012
    GATT_PROXY_SET = 0x8013
    GATT_PROXY_STATUS = 0x8014
    HEARBEAT_PUBLICATION_GET = 0x8038
    HEARBEAT_PUBLICATION_SET = 0x8039
    HEARBEAT_PUBLICATION_STATUS = 0x06
    HEARBEAT_SUBSCRIPTION_GET = 0x803A
    HEARBEAT_SUBSCRIPTION_SET = 0x803B
    HEARBEAT_SUBSCRIPTION_STATUS = 0x803C
    KEY_REFRESH_PHASE_GET = 0x8015
    KEY_REFRESH_PHASE_SET = 0x8016
    KEY_REFRESH_PHASE_STATUS = 0x8017
    LOW_POWER_NODE_POLLTIMEOUT_GET = 0x802D
    LOW_POWER_NODE_POLLTIMEOUT_STATUS = 0x802E
    MODEL_APP_BIND = 0x803D
    MODEL_APP_STATUS = 0x803E
    MODEL_APP_UNBIND = 0x803F
    MODEL_PUBLICATION_GET = 0x8018
    MODEL_PUBLICATION_SET = 0x03
    MODEL_PUBLICATION_STATUS = 0x8019
    MODEL_PUBLICATION_VIRTUAL_ADDRESS_SET = 0x801A
    MODEL_SUBSCRIPTION_ADD = 0x801B
    MODEL_SUBSCRIPTION_DELETE = 0x801C
    MODEL_SUBSCRIPTION_DELETE_ALL = 0x801D
    MODEL_SUBSCRIPTION_OVERWRITE = 0x801E
    MODEL_SUBSCRIPTION_STATUS = 0x801F
    MODEL_SUBSCRIPTION_VIRTUAL_ADDRESS_ADD = 0x8020
    MODEL_SUBSCRIPTION_VIRTUAL_ADDRESS_DELETE = 0x8021
    MODEL_SUBSCRIPTION_VIRTUAL_ADDRESS_OVERWRITE = 0x8022
    NETKEY_ADD = 0x8040
    NETKEY_DELETE = 0x8041
    NETKEY_GET = 0x8042
    NETKEY_LIST = 0x8043
    NETKEY_STATUS = 0x8044
    NETKEY_UPDATE = 0x8045
    NETWORK_TRANSMIT_GET = 0x8023
    NETWORK_TRANSMIT_SET = 0x8024
    NETWORK_TRANSMIT_STATUS = 0x8025
    NODE_IDENTITY_GET = 0x8046
    NODE_IDENTITY_SET = 0x8047
    NODE_IDENTITY_STATUS = 0x8048
    NODE_RESET = 0x8049
    NODE_RESET_STATUS = 0x804A
    RELAY_GET = 0x8026
    RELAY_SET = 0x8027
    RELAY_STATUS = 0x8028
    SIG_MODEL_APP_GET = 0x804B
    SIG_MODEL_APP_LIST = 0x804C
    SIG_MODEL_SUBSCRIPTION_GET = 0x8029
    SIG_MODEL_SUBSCRIPTION_LIST = 0x802A
    VENDOR_MODEL_APP_GET = 0x804D
    VENDOR_MODEL_APP_LIST = 0x804E
    VENDOR_MODEL_SUBSCRIPTION_GET = 0x802B
    VENDOR_MODEL_SUBSCRIPTION_LIST = 0x802C


# fmt: off
ConfigMessage = Struct(
    "opcode" / Opcode(ConfigOpcode),
    "params" / Switch(
        this.opcode,
        {
            ConfigOpcode.APPKEY_ADD: ConfigAppKeyAdd,
            ConfigOpcode.APPKEY_DELETE: ConfigAppKeyDelete,
            ConfigOpcode.APPKEY_GET: ConfigAppKeyGet,
            ConfigOpcode.APPKEY_LIST: ConfigAppKeyList,
            ConfigOpcode.APPKEY_STATUS: ConfigAppKeyStatus,
            ConfigOpcode.APPKEY_UPDATE: ConfigAppKeyUpdate,
            ConfigOpcode.BEACON_GET: ConfigBeaconGet,
            ConfigOpcode.BEACON_SET: ConfigBeaconSet,
            ConfigOpcode.BEACON_STATUS: ConfigBeaconStatus,
            ConfigOpcode.COMPOSITION_DATA_GET: ConfigCompositionDataGet,
            ConfigOpcode.COMPOSITION_DATA_STATUS: ConfigCompositionDataStatus,
            ConfigOpcode.DEFAULT_TTL_GET: ConfigDefaultTTLGet,
            ConfigOpcode.DEFAULT_TTL_SET: ConfigDefaultTTLSet,
            ConfigOpcode.DEFAULT_TTL_STATUS: ConfigDefaultTTLStatus,
            ConfigOpcode.FRIEND_GET: ConfigFriendGet,
            ConfigOpcode.FRIEND_SET: ConfigFriendSet,
            ConfigOpcode.FRIEND_STATUS: ConfigFriendStatus,
            ConfigOpcode.GATT_PROXY_GET: ConfigGATTProxyGet,
            ConfigOpcode.GATT_PROXY_SET: ConfigGATTProxySet,
            ConfigOpcode.GATT_PROXY_STATUS: ConfigGATTProxyStatus,
            ConfigOpcode.HEARBEAT_PUBLICATION_GET: ConfigHeartbeatPublicationGet,
            ConfigOpcode.HEARBEAT_PUBLICATION_SET: ConfigHeartbeatPublicationSet,
            ConfigOpcode.HEARBEAT_PUBLICATION_STATUS: ConfigHeartbeatPublicationStatus,
            ConfigOpcode.HEARBEAT_SUBSCRIPTION_GET: ConfigHeartbeatSubscriptionGet,
            ConfigOpcode.HEARBEAT_SUBSCRIPTION_SET: ConfigHeartbeatSubscriptionSet,
            ConfigOpcode.HEARBEAT_SUBSCRIPTION_STATUS: ConfigHeartbeatSubscriptionStatus,
            ConfigOpcode.KEY_REFRESH_PHASE_GET: ConfigKeyRefreshPhaseGet,
            ConfigOpcode.KEY_REFRESH_PHASE_SET: ConfigKeyRefreshPhaseSet,
            ConfigOpcode.KEY_REFRESH_PHASE_STATUS: ConfigKeyRefreshPhaseStatus,
            ConfigOpcode.LOW_POWER_NODE_POLLTIMEOUT_GET: ConfigLowPowerNodePollTimeoutGet,
            ConfigOpcode.LOW_POWER_NODE_POLLTIMEOUT_STATUS: ConfigLowPowerNodePollTimeoutStatus,
            ConfigOpcode.MODEL_APP_BIND: ConfigModelAppBind,
            ConfigOpcode.MODEL_APP_STATUS: ConfigModelAppStatus,
            ConfigOpcode.MODEL_APP_UNBIND: ConfigModelAppUnbind,
            ConfigOpcode.MODEL_PUBLICATION_GET: ConfigModelPublicationGet,
            ConfigOpcode.MODEL_PUBLICATION_SET: ConfigModelPublicationSet,
            ConfigOpcode.MODEL_PUBLICATION_STATUS: ConfigModelPublicationStatus,
            ConfigOpcode.MODEL_PUBLICATION_VIRTUAL_ADDRESS_SET: ConfigModelPublicationVASet,
            ConfigOpcode.MODEL_SUBSCRIPTION_ADD: ConfigModelSubscriptionAdd,
            ConfigOpcode.MODEL_SUBSCRIPTION_DELETE: ConfigModelSubscriptionDelete,
            ConfigOpcode.MODEL_SUBSCRIPTION_DELETE_ALL: ConfigModelSubscriptionDeleteAll,
            ConfigOpcode.MODEL_SUBSCRIPTION_OVERWRITE: ConfigModelSubscriptionOverwrite,
            ConfigOpcode.MODEL_SUBSCRIPTION_STATUS: ConfigModelSubscriptionStatus,
            ConfigOpcode.MODEL_SUBSCRIPTION_VIRTUAL_ADDRESS_ADD: ConfigModelSubscriptionVAAdd,
            ConfigOpcode.MODEL_SUBSCRIPTION_VIRTUAL_ADDRESS_DELETE: ConfigModelSubscriptionVADelete,
            ConfigOpcode.MODEL_SUBSCRIPTION_VIRTUAL_ADDRESS_OVERWRITE: ConfigModelSubscriptionVAOverwrite,
            ConfigOpcode.NETKEY_ADD: ConfigNetKeyAdd,
            ConfigOpcode.NETKEY_DELETE: ConfigNetKeyDelete,
            ConfigOpcode.NETKEY_GET: ConfigNetKeyGet,
            ConfigOpcode.NETKEY_LIST: ConfigNetKeyList,
            ConfigOpcode.NETKEY_STATUS: ConfigNetKeyStatus,
            ConfigOpcode.NETKEY_UPDATE: ConfigNetKeyUpdate,
            ConfigOpcode.NETWORK_TRANSMIT_GET: ConfigNetworkTransmitGet,
            ConfigOpcode.NETWORK_TRANSMIT_SET: ConfigNetworkTransmitSet,
            ConfigOpcode.NETWORK_TRANSMIT_STATUS: ConfigNetworkTransmitStatus,
            ConfigOpcode.NODE_IDENTITY_GET: ConfigNodeIdentityGet,
            ConfigOpcode.NODE_IDENTITY_SET: ConfigNodeIdentitySet,
            ConfigOpcode.NODE_IDENTITY_STATUS: ConfigNodeIdentityStatus,
            ConfigOpcode.NODE_RESET: ConfigNodeReset,
            ConfigOpcode.NODE_RESET_STATUS: ConfigNodeResetStatus,
            ConfigOpcode.RELAY_GET: ConfigRelayGet,
            ConfigOpcode.RELAY_SET: ConfigRelaySet,
            ConfigOpcode.RELAY_STATUS: ConfigRelayStatus,
            ConfigOpcode.SIG_MODEL_APP_GET: ConfigSIGModelAppGet,
            ConfigOpcode.SIG_MODEL_APP_LIST: ConfigSIGModelAppList,
            ConfigOpcode.SIG_MODEL_SUBSCRIPTION_GET: ConfigSIGModelSubscriptionGet,
            ConfigOpcode.SIG_MODEL_SUBSCRIPTION_LIST: ConfigSIGModelSubscriptionList,
            ConfigOpcode.VENDOR_MODEL_APP_GET: ConfigVendorModelAppGet,
            ConfigOpcode.VENDOR_MODEL_APP_LIST: ConfigVendorModelAppList,
            ConfigOpcode.VENDOR_MODEL_SUBSCRIPTION_GET: ConfigVendorModelSubscriptionGet,
            ConfigOpcode.VENDOR_MODEL_SUBSCRIPTION_LIST: ConfigVendorModelSubscriptionList,
        }
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest
from construct import SelectError, ValidationError

from bluetooth_mesh.messages.scene import SceneMessage, SceneOpcode

invalid = [
    # fmt: off
    pytest.param(
        SelectError,
        SceneOpcode.SCENE_RECALL,
        dict(scene_number=0, tid=30),
        id="SCENE_RECALL_minimal_with_prohibited_scene_number"),
    pytest.param(
        SelectError,
        SceneOpcode.SCENE_RECALL,
        dict(scene_number=0, tid=30, transition_time=30000, delay=0.3),
        id="SCENE_RECALL_with_transition_with_prohibited_scene_number"),
    pytest.param(
        SelectError,
        SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED,
        dict(scene_number=0, tid=30),
        id="SCENE_RECALL_UNACKNOWLEDGED_minimal_with_prohibited_scene_number"),
    pytest.param(
        SelectError,
        SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED,
        dict(scene_number=0, tid=30, transition_time=30000, delay=0.3),
        id="SCENE_RECALL_UNACKNOWLEDGED_with_transition_with_prohibited_scene_number"),
    pytest.param(
        ValidationError,
        SceneOpcode.SCENE_STORE,
        dict(scene_number=0),
        id="SCENE_STORE_with_prohibited_scene_number"),
    pytest.param(
        ValidationError,
        SceneOpcode.SCENE_STORE_UNACKNOWLEDGED,
        dict(scene_number=0),
        id="SCENE_STORE_UNACKNOWLEDGED_with_prohibited_scene_number")
    # fmt: on
]

valid = [
    # fmt: off
    pytest.param(
        b'\x82\x41',
        SceneOpcode.SCENE_GET,
        dict(),
        id="SCENE_GET"),
    pytest.param(
        b'\x82\x42\x01\x00\x1e',
        SceneOpcode.SCENE_RECALL,
        dict(scene_number=1, tid=30),
        id="SCENE_RECALL_minimal"),
    pytest.param(
        b'\x82\x43\x01\x00\x1e',
        SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED,
        dict(scene_number=1, tid=30),
        id="SCENE_RECALL_UNACKNOWLEDGED_minimal"),
    pytest.param(
        b'\x82\x42\x01\x00\x1e\xf2\x3c',
        SceneOpcode.SCENE_RECALL,
        dict(scene_number=1, tid=30, transition_time=30000, delay=0.3),
        id="SCENE_RECALL_with_transition"),
    pytest.param(
        b'\x82\x43\x01\x00\x1e\xf2\x3c',
        SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED,
        dict(scene_number=1, tid=30, transition_time=30000, delay=0.3),
        id="SCENE_RECALL_UNACKNOWLEDGED_with_transition"),
    pytest.param(
        b'\x5e\x00\x01\x00',
        SceneOpcode.SCENE_STATUS,
        dict(status_code=0, current_scene=1),
        id="SCENE_STATUS_minimal"),
    pytest.param(
        b'\x5e\x00\x01\x00\x02\x00\xf2',
        SceneOpcode.SCENE_STATUS,
        dict(status_code=0, current_scene=1, target_scene=2, remaining_time=30000),
        id="SCENE_STATUS_with_target_scene"),
    pytest.param(
        b'\x82\x44',
        SceneOpcode.SCENE_REGISTER_GET,
        dict(),
        id="SCENE_REGISTER_GET"),
    pytest.param(
        b'\x82\x45\x00\x01\x00\x01\x00\x02\x00' + 14 * b'\x00\x00',
        SceneOpcode.SCENE_REGISTER_STATUS,
        dict(status_code=0, current_scene=1, scenes=[1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        id="SCENE_REGISTER_STATUS"),
    pytest.param(
        b'\x82\x46\x01\x00',
        SceneOpcode.SCENE_STORE,
        dict(scene_number=1),
        id="SCENE_STORE"),
    pytest.param(
        b'\x82\x47\x01\x00',
        SceneOpcode.SCENE_STORE_UNACKNOWLEDGED,
        dict(scene_number=1),
        id="SCENE_STORE_UNACKNOWLEDGED"),
    pytest.param(
        b'\x82\x9e\x01\x00',
        SceneOpcode.SCENE_DELETE,
        dict(scene_number=1),
        id="SCENE_DELETE"),
    pytest.param(
        b'\x82\x9f\x01\x00',
        SceneOpcode.SCENE_DELETE_UNACKNOWLEDGED,
        dict(scene_number=1),
        id="SCENE_DELETE_UNACKNOWLEDGED"),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert SceneMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert SceneMessage.build(dict(opcode=opcode, params=data)) == encoded


@pytest.mark.parametrize("exception,opcode,data", invalid)
def test_build_valid(exception, opcode, data):
    with pytest.raises(exception):
        SceneMessage.build(dict(opcode=opcode, params=data))



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from datetime import timedelta
from enum import IntEnum

from construct import (
    Adapter,
    Const,
    Default,
    Flag,
    Int8ul,
    Int16ul,
    Int24ul,
    Struct,
    Switch,
    this,
)

from bluetooth_mesh.messages.util import EnumAdapter, Opcode


class LightExtendedControllerOpcode(IntEnum):
    OPCODE = 0xF63601


class LightExtendedControllerSubOpcode(IntEnum):
    PROPERTY_GET = 0x00
    PROPERTY_SET = 0x01
    PROPERTY_SET_UNACKNOWLEDGED = 0x01
    PROPERTY_STATUS = 0x03
    BULK_LC_PROPERTY_SET = 0x04
    BULK_LC_PROPERTY_STATUS = 0x05
    BULK_LEC_PROPERTY_SET = 0x06
    BULK_LEC_PROPERTY_STATUS = 0x06
    SYNC_INTEGRAL_GET = 0x08
    SYNC_INTEGRAL_STATUS = 0x09


class LightExtendedControllerProperty(IntEnum):
    AUTO_RESUME_MODE = 0xFF71
    AUTO_RESUME_TIMER = 0xFF72


class AutoResumeTimerAdapter(Adapter):
    def _decode(self, obj, context, path):
        return timedelta(milliseconds=obj)

    def _encode(self, obj, context, path):
        return int(obj.total_seconds() * 1000)


# fmt: off
LightExtendedControllerPropertyValue = Switch(
    this.id,
    {
        LightExtendedControllerProperty.AUTO_RESUME_MODE: Flag,
        LightExtendedControllerProperty.AUTO_RESUME_TIMER: AutoResumeTimerAdapter(Int24ul),
    }
)

LightExtendedControllerPropertyGet = Struct(
    "id" / EnumAdapter(Int16ul, LightExtendedControllerProperty),
)

LightExtendedControllerPropertySet = Struct(
    "id" / EnumAdapter(Int16ul, LightExtendedControllerProperty),
    "value" / LightExtendedControllerPropertyValue,
)

LightExtendedControllerPropertyStatus = Struct(
    "id" / EnumAdapter(Int16ul, LightExtendedControllerProperty),
    "value" / LightExtendedControllerPropertyValue,
)

LightExtendedControllerSyncIntegralStatus = Struct(
    "sync_integral" / Int16ul,
)

LightExtendedControllerPayload = Struct(
    "subopcode" / EnumAdapter(Int8ul, LightExtendedControllerSubOpcode),
    "payload" / Default(Switch(
        this.subopcode,
        {
            LightExtendedControllerSubOpcode.PROPERTY_GET: LightExtendedControllerPropertyGet,
            LightExtendedControllerSubOpcode.PROPERTY_SET: LightExtendedControllerPropertySet,
            LightExtendedControllerSubOpcode.PROPERTY_SET_UNACKNOWLEDGED: LightExtendedControllerPropertySet,
            LightExtendedControllerSubOpcode.PROPERTY_STATUS: LightExtendedControllerPropertyStatus,
            LightExtendedControllerSubOpcode.SYNC_INTEGRAL_STATUS: LightExtendedControllerSyncIntegralStatus,
        },
    ), None)
)

LightExtendedControllerMessage = Struct(
    "opcode" / Const(LightExtendedControllerOpcode.OPCODE, Opcode(LightExtendedControllerOpcode)),
    "params" / LightExtendedControllerPayload
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    Flag,
    GreedyBytes,
    Int8ul,
    Int16ul,
    Struct,
    Switch,
    Padding,
    this,
)

from bluetooth_mesh.messages.util import EnumAdapter


class FdMessageType(IntEnum):
    NETKEY = 0
    APPKEY = 1


# fmt: off
NetKeyMessageElements = Struct(
    "net_index" / Int16ul,
    "remote" / Flag,
    Padding(17)
)

AppKeyMessageElements = Struct(
    "app_index" / Int16ul,
    "destination" / Int16ul,
    "label" / Int8ul,
    Padding(15)
)


FdMessage = Struct(
    "element" / Int8ul,
    "source" / Int16ul,
    "type" / EnumAdapter(Int8ul, FdMessageType),
    "params" / Switch(
        this.type,
        {
            FdMessageType.NETKEY: NetKeyMessageElements,
            FdMessageType.APPKEY: AppKeyMessageElements,
        },
    ),
    "data" / GreedyBytes,
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest
from asynctest import CoroutineMock

from bluetooth_mesh.utils import Signal


@pytest.mark.asyncio
async def test_signal():
    cb1 = CoroutineMock()
    cb2 = CoroutineMock()

    ts = Signal()
    ts.connect(cb1)
    ts.connect(cb2)

    await ts.emit("test msg")
    cb1.assert_called_with("test msg")
    cb2.assert_called_with("test msg")


@pytest.mark.asyncio
async def test_signal_keywords():
    cb1 = CoroutineMock()
    cb2 = CoroutineMock()

    ts = Signal()
    ts.connect(cb1)
    ts.connect(cb2)

    await ts.emit(test_msg="test msg")
    cb1.assert_called_with(test_msg="test msg")
    cb2.assert_called_with(test_msg="test msg")


@pytest.mark.asyncio
async def test_signal_cb_exception():
    cb_ok = CoroutineMock()

    async def cb_exception(*args, **kwargs):
        raise Exception

    ts = Signal()
    ts.connect(cb_ok)
    ts.connect(cb_exception)

    await ts.emit("test msg")
    cb_ok.assert_called_with("test msg")


@pytest.mark.asyncio
async def test_signal_sync_async_cb():
    class Mock:
        def __init__(self):
            self.async_ret = None
            self.sync_ret = None

        async def async_cb(self, *args, **kwargs):
            self.async_ret = args

        def sync_cb(self, *args, **kwargs):
            self.sync_ret = args

    mocker = Mock()

    ts = Signal()
    ts.connect(mocker.async_cb)
    ts.connect(mocker.sync_cb)

    await ts.emit("test msg")
    assert mocker.async_ret == ("test msg",)
    assert mocker.sync_ret == ("test msg",)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from pytest import fixture

from bluetooth_mesh.crypto import k1, k2, k3, k4, s1


@fixture
def app_key():
    return bytes.fromhex("3216d1509884b533248541792b877f98")


@fixture
def net_key():
    return bytes.fromhex("f7a2a44f8e8a8029064f173ddc1e2b00")


@fixture
def dev_key():
    return bytes.fromhex("37c612c4a2d337cb7b98355531b3617f")


def test_s1():
    s = s1(b"test")
    assert s == bytes.fromhex("b73cefbd641ef2ea598c2b6efb62f79c")


def test_k1(app_key):
    N = app_key
    SALT = bytes.fromhex("2ba14ffa0df84a2831938d57d276cab4")
    P = bytes.fromhex("5a09d60797eeb4478aada59db3352a0d")

    k = k1(N, SALT, P)

    assert k == bytes.fromhex("f6ed15a8934afbe7d83e8dcb57fcf5d7")


def test_k2_master(net_key):
    N = net_key
    P = b"\x00"

    n, e, p = k2(N, P)

    assert n == 0x7F
    assert e == bytes.fromhex("9f589181a0f50de73c8070c7a6d27f46")
    assert p == bytes.fromhex("4c715bd4a64b938f99b453351653124f")


def test_k3(net_key):
    N = net_key

    n = k3(N)

    assert n == bytes.fromhex("ff046958233db014")


def test_k4(app_key):
    N = app_key

    a = k4(N)

    assert a == 0x38



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#

# pylint: disable=C0103

from functools import lru_cache

import bitstring
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import cmac
from cryptography.hazmat.primitives.ciphers import Cipher, aead, algorithms, modes


def aes_cmac(k, m):
    c = cmac.CMAC(algorithms.AES(k), backend=default_backend())
    c.update(m)
    return c.finalize()


def aes_ccm_encrypt(k, n, m, a=b"", tag_length=4):
    c = aead.AESCCM(k, tag_length)
    return c.encrypt(n, m, a)


def aes_ccm_decrypt(k, n, m, a=b"", tag_length=4):
    c = aead.AESCCM(k, tag_length)
    return c.decrypt(n, m, a)


def aes_ecb(k, m):
    c = Cipher(algorithms.AES(k), modes.ECB(), backend=default_backend())
    e = c.encryptor()
    return e.update(m) + e.finalize()


def s1(M):
    ZERO = bytes([0] * 16)
    return aes_cmac(ZERO, M)


def k1(N, SALT, P):
    T = aes_cmac(SALT, N)
    return aes_cmac(T, P)


def k2(N, P):
    SALT = s1(b"smk2")
    T = aes_cmac(SALT, N)
    T0 = b""
    T1 = aes_cmac(T, T0 + P + b"\x01")
    T2 = aes_cmac(T, T1 + P + b"\x02")
    T3 = aes_cmac(T, T2 + P + b"\x03")

    k = (T1 + T2 + T3)[-33:]

    n, e, p = bitstring.BitString(k).unpack("pad:1, uint:7, bits:128, bits:128")

    return n, e.bytes, p.bytes


def k3(N):
    SALT = s1(b"smk3")
    T = aes_cmac(SALT, N)
    return aes_cmac(T, b"id64\x01")[-8:]


def k4(N):
    SALT = s1(b"smk4")
    T = aes_cmac(SALT, N)

    k = aes_cmac(T, b"id6\x01")[-1:]

    (aid,) = bitstring.BitString(k).unpack("pad:2, uint:6")

    return aid


class Key:
    def __init__(self, key):
        self.bytes = key

    def __str__(self):
        return "<%s: %s>" % (type(self).__name__, self.bytes.hex())


class ApplicationKey(Key):
    @property
    @lru_cache(maxsize=1)
    def aid(self):
        return k4(self.bytes)


class DeviceKey(Key):
    @property
    def aid(self):
        return 0


class NetworkKey(Key):
    @property
    @lru_cache(maxsize=1)
    def network_id(self):
        return k3(self.bytes)

    @property
    @lru_cache(maxsize=1)
    def encryption_keys(self):
        return k2(self.bytes, b"\x00")

    @property
    @lru_cache(maxsize=1)
    def identity_key(self):
        return k1(self.bytes, s1(b"nkik"), b"id128\x01")

    @property
    @lru_cache(maxsize=1)
    def beacon_key(self):
        return k1(self.bytes, s1(b"nkbk"), b"id128\x01")
